; Mini-threads library.
; x86 NASM assembly code
; Orion Sky Lawlor, olawlor@acm.org, 2006/01/29 (Public Domain)
;  Compile into a Linux .o file with "nasm -f elf threads.s -o threads.o"
;    or a Windows .obj file with "nasm -f win32 threads.s -o threads.obj"

SECTION .text

; void mth_swap(mth_thread *cur_th,mth_thread *new_th)
; Save the state of the current thread 
;  and switch to this other thread stack.
; 
;  
; A suspended thread's stack looks like this:
;  new_th    [ or for a new thread: f argument   ]
;  cur_th    [ or for a new thread: post-f code  ]
;  return pc [ or for a new thread: pointer to f ]
;  ebp
;  esi
;  edi  <- saved thread's esp points here
GLOBAL _mth_swap
_mth_swap: ; (Windows demands name with underscore)
GLOBAL mth_swap ; (mth_swap is callable from outside)
mth_swap:
	mov eax,[esp+4] ; points to cur_th struct
	mov ecx,[esp+8] ; points to new_th struct
	; Push all saved processor registers
	push ebp
	push esi
	push edi
	; Save old thread's stack pointer
	mov [eax],esp  ; cur_th->stackPtr = esp
	; Load new thread's stack pointer
	mov esp,[ecx]  ; esp = new_th->stackPtr
	; Restore all saved processor registers
	pop edi
	pop esi
	pop ebp
	ret  ; Jump back to new thread


; void mth_create(mth_thread *n,fn f,void *fn_arg)
; Sets up a new thread stack.
;
; Tweaks this new thread so that when swapped to, 
;  the "ret" at the end of mth_swap will jump to 
;  the function f, passing argument fn_arg.
; 
; This function could just as easily be written in C/C++.
;
GLOBAL _mth_create
_mth_create:
GLOBAL mth_create ; (mth_create is callable from outside)
mth_create:
	mov edx,[esp+4] ; n
	mov ecx,[edx] ; top of n's stack
	mov eax,[esp+12] ; push fn_arg onto new thread's stack
	mov [ecx-4],eax
	mov long [ecx-8],0x01  ; Address for f to return to
	mov eax,[esp+8] ; f itself
	mov [ecx-12],eax
	mov long [ecx-16],0x12 ; initial ebp inside f
	mov long [ecx-20],0x23 ; initial esi inside f
	mov long [ecx-24],0x34 ; initial edi inside f
	sub ecx,24
	mov [edx],ecx ; save updated stack pointer
	ret
